#!/usr/bin/perl
# prepare microarray data for use in EMBER w/ discretized behaviors

use Getopt::Long;
$|++;

#
# command line arguments
#
$options = "Usage: ./PreProcess_Expresssion_Data.pl <OPTIONS>
	-i  list of microarray files (required)
	    files should be named <condition><replicate>.txt; replicate should start at 1
	    format: <id> <value>
	-c  list of comparisons (required)
	    format: <condition1> <condition2>
	-a  annotation file (required)
	    format: <id> <gene name> <chr> <start> <end> <strand>
	-o  output (required)
	-p  percentile threshold (number between 0 and 1, default 0.63)
	    this value is used as follows: all data are concatenated into one list,
	    and the pth percentile of that list is taken as the threshold.
	    Then, a probe is removed if its value is less than the threshold in ALL conditions.
	    p=1.0 means all probes are retained, p=0.0 means none are
	-l  log-transform the data (y or n, default n)
	-v  verbose (print progress to stdout: y or n, default y)
\n";

$i = "";
$c = "";
$a = "";
$o = "";
$p = 0.63;
$l = "n";
$v = "y";

GetOptions('i=s' => \$i,
	   'c=s' => \$c,
	   'a=s' => \$a,
	   'o=s' => \$o,
	   'p=f' => \$p,
	   'l=s' => \$l,
	   'v=s' => \$v,
) || die "\n$options\n";

if( $i eq "" ){
	print "\nError: set a value for -i\n\n$options";
	exit;
}
if( $c eq "" ){
	print "\nError: set a value for -c\n\n$options";
	exit;
}
if( $a eq "" ){
	print "\nError: set a value for -a\n\n$options";
	exit;
}
if( $o eq "" ){
	print "\nError: set a value for -o\n\n$options";
	exit;
}

if( $p < 0 || $p > 1 ){
	print "\nError: choose -p to be between 0 and 1\n\n$options";
	exit;
}
if( $l ne "y" && $l ne "n" ){
	print "\nError: choose -l to be y or n\n\n$options";
	exit;
}
if( $v ne "y" && $v ne "n" ){
	print "\nError: choose -v to be y or n\n\n$options";
	exit;
}

#################################################
# SECTION 1 - read in file names and conditions #
#################################################

#
# read in list of files and list of comparisons
# arrays: @files, @names, @comps
#
open(IN,"$i") || die "Error: can't open file $i\n";
@files = ();
@names = ();
while($line = <IN>){
	chomp($line);
	push(@files, $line);
	@parts = split('/',$line);
	($base, $suff) = split('\.txt', $parts[$#parts]);
	push(@names, $base);
}
close(IN);
open(IN,"$c") || die "Error: can't open file $c\n";
@comps = ();
while($line = <IN>){
	chomp($line);
	@parts = split(' ',$line);
	if( $#parts != 1 ){
		print "Error: file $c does not have 2 columns\n";
		exit;
	}
	$info1 = {};
	$info1->{name} = $parts[0];
	$info1->{namei} = 0;
	$info2 = {};
	$info2->{name} = $parts[1];
	$info2->{namei} = 0;
	push(@comps, [($info1, $info2)] );
}
close(IN);

#
# count how many conditions we have, and how many replicates of each
# arrays: @conds
#
@conds = ();
for($i=0; $i<= $#comps; $i++){
	for($j=0; $j< 2; $j++){
		$ok = 1;
		$k=-1;
		$end = $#conds;
		while($k< $end && $ok == 1){
			$k++;
			if( $conds[$k]->{name} eq $comps[$i][$j]->{name} ){
				$ok = 0;
			}
		}
		if( $ok == 1 ){
			@empty1 = ();
			$info = {};
			$info->{name} = $comps[$i][$j]->{name};
			$info->{reps} = [@empty1];
			$info->{nreps} = 0;
			push(@conds, $info);
			$comps[$i][$j]->{namei} = $#conds;
		}
		else{
			$comps[$i][$j]->{namei} = $k;
		}
	}
}
for($i=0; $i<= $#conds; $i++){
	for($j=0; $j<= $#names; $j++){
		if( $names[$j] =~ /$conds[$i]->{name}/ ){
			push(@{$conds[$i]->{reps}}, $j);
			$conds[$i]->{nreps}++;
		}
	}
}

#
# print summary to stdout
#
if( $v eq "y" ){
	printf("\nReading in %i conditions, %i total files:\n", $#conds+1, $#files+1);
	for($i=0; $i<= $#conds; $i++){
		printf("   %s, %i replicates\n", $conds[$i]->{name}, $conds[$i]->{nreps});
	}
	printf("\nMaking %i comparisons:\n", $#comps+1);
	for($i=0; $i<= $#comps; $i++){
		printf("   %s vs %s\n", $comps[$i][0]->{name}, $comps[$i][1]->{name});
	}
}

###########################################################
# SECTION 2 - read in annotation file and microarray data #
###########################################################

if( $v eq "y" ){
	print "\nReading in annotation file and data\n";
}

#
# read in probe annotations
# arrays: @probes
#
open(IN,"$a") || die "Error: can't open file $a\n";
@vals = ();
while($line = <IN>){
	chomp($line);
	@parts = split(' ',$line);
	if( $#parts != 5 ){
		print "Error: file $a does not have 6 columns\n";
		exit;
	}
	@empty1 = ();
	@empty2 = ();
	$info = {};
	$info->{id} = $parts[0];
	$info->{name} = $parts[1];
	$info->{chr} = $parts[2];
	$info->{start} = $parts[3];
	$info->{end} = $parts[4];
	$info->{strand} = $parts[5];
	$info->{vals} = [@empty1];
	$info->{classes} = [@empty2];
	$info->{use} = 0;
	push(@vals, $info);
}
close(IN);
@probes = sort{ $a->{id} cmp $b->{id} } @vals;

#
# read in each file, sort by ids, and add to @probes
# record average for each file
# arrays: @averages
#
@averages = ();
for($i=0; $i<= $#files; $i++){
	#
	# read in data
	#
	if( $v eq "y" ){
		printf("\r   %i of %i files", $i+1, $#files+1);
	}
	open(IN,"$files[$i]") || die "Error: can't open $files[$i]\n";
	@vals = ();
	while($line = <IN>){
		chomp($line);
		@parts = split(' ',$line);
		if( $#parts == 1 ){
			push(@vals, [@parts] );
		}
	}
	close(IN);
	@svals = sort{ $a->[0] cmp $b->[0] } @vals;

	#
	# add to @probes
	#
	$end = $#svals;
	$j=0;
	$avg=0;
	for($k=0; $k<= $#probes; $k++){
		while( $probes[$k]->{id} gt $svals[$j][0] && $j<= $end ){$j++;}
		if( $probes[$k]->{id} eq $svals[$j][0] ){
			if( $l eq "n" ){ push(@{$probes[$k]->{vals}}, $svals[$j][1]);}
			if( $l eq "y" ){ push(@{$probes[$k]->{vals}}, log($svals[$j][1])/log(2) );}
			$j++;
		}
		else{
			$j=0;
		#	print "\nError: did not find a value for probe $probes[$k]->{id} in $files[$i]\n";
		#	exit;
		}
	}
	
	#
	# compute average
	# 
	$avg = 0;
	for($k=0; $k<= $#probes; $k++){
		$avg+= $probes[$k]->{vals}[$i];
	}
	$avg/= ($#probes+1);
	push(@averages, $avg);
}

if( $v eq "y" ){
	printf("\n   using %i probes\n", $#probes+1);
}

##################################################################
# SECTION 3 - determine threshold and label probes to be removed #
##################################################################

if( $v eq "y" ){
	printf("\nDetermining threshold\n");
}
#
# find threshold
# variables: $thresh
#
@allvals = ();
for($i=0; $i<= $#probes; $i++){
	for($j=0; $j<= $#files; $j++){
		push(@allvals, ($probes[$i]->{vals}[$j])/$averages[$j]);
	}
}

@sallvals = sort{ $a <=> $b } @allvals;

$end = (1-$p)*$#sallvals;
$i = sprintf("%i", $end);
$thresh = $sallvals[$i];

#
# label probes as on or off
#
$totuse = 0;
for($i=0; $i<= $#probes; $i++){
	$j=0;
	while($j<= $#files && $probes[$i]->{use} == 0 ){
		if( ($probes[$i]->{vals}[$j])/$averages[$j] > $thresh ){
			$probes[$i]->{use} = 1;
			$totuse++;
		}
		$j++;
	}
}
if( $v eq "y" ){
	printf("   %i probes retained for analysis\n", $totuse);
}

#############################################
# SECTION 4 - make discrete classifications #
#############################################

if( $v eq "y" ){
	printf("\nClassifying probes\n");
}
#
# for each probe,
#    for each comparison,
#       compute avg and stdev of first and second conditions
#       apply rules for classification
#
for($i=0; $i<= $#probes; $i++){
	for($j=0; $j<= $#comps; $j++){
		$avg = (0,0);
		$sd = (0,0);
		for($k=0; $k< 2; $k++){
			$condi = $comps[$j][$k]->{namei};
			$avg[$k] = 0;
			$sd[$k] = 0;
			for($l=0; $l< $conds[$condi]->{nreps}; $l++){
				$filei = $conds[$condi]->{reps}[$l];
				$avg[$k]+= $probes[$i]->{vals}[$filei];
			}
			$avg[$k] /= $conds[$condi]->{nreps};
			for($l=0; $l< $conds[$condi]->{nreps}; $l++){
				$filei = $conds[$condi]->{reps}[$l];
				$sd[$k]+= ($avg[$k]-$probes[$i]->{vals}[$filei])**2;
			}
			$sd[$k] = sqrt($sd[$k]/($conds[$condi]->{nreps}));
		}

		$s = $sd[0] + $sd[1];
		# ++ category
		if( $avg[1]-$avg[0] > 3*$s ){
			push( @{$probes[$i]->{classes}}, 2);
		}
		# + category
		elsif( $avg[1]-$avg[0] > $s ){
			push(@{$probes[$i]->{classes}}, 1);
		}
		# -- category
		elsif( $avg[0]-$avg[1] > 3*$s ){
			push(@{$probes[$i]->{classes}}, -2);
		}
		# 0 category
		elsif( $avg[0]-$avg[1] > $s ){
			push(@{$probes[$i]->{classes}}, -1);
		}
		else{
			push(@{$probes[$i]->{classes}}, 0);
		}
	}
}

############################
# SECTION 5 - print result #
############################

if( $v eq "y" ){
	printf("\nPrinting result\n\n");
}
open(OUT,">$o");
$delim = " ";
for($i=0; $i<= $#probes; $i++){
	printf OUT ("%s%s%s%s%s%s%s%s%s%s%s", $probes[$i]->{id}, $delim, $probes[$i]->{name}, $delim, $probes[$i]->{chr}, $delim, $probes[$i]->{start}, $delim, $probes[$i]->{end}, $delim, $probes[$i]->{strand} );
	if( $probes[$i]->{use} == 1 ){
		for($j=0; $j<= $#comps; $j++){
			printf OUT ("%s%i", $delim, $probes[$i]->{classes}[$j]);
		}
		print OUT "\n";
	}
	else{
		print OUT "$delim-1\n";
	}
}
close(OUT);








