
/*
 * various functions to be performed on arrays, files, etc
 */

// functions on files
int count_lines( FILE *fin, char *filename);
int longest_line( FILE *fin, char *filename);
char *get_title( FILE *fin, char *filename);
int count_nclines( FILE *fin, char *filename);
char **read_list( FILE *fin, char *filename, int *len, char *dir );

//functions on strings
int check_ext( char *filename, char *ext );
int check_comment( char *field );
char *getdir( char *filename );
int numfields( char *string, char *delims );
char *evena( char *name, int maxlen );
char *eveni( int val , char *name, int maxlen );
char *evenf( double val , char *name, int maxlen );
char *evene( double val , char *name, int maxlen );
void even( char *name, int maxlen );
char **split( const char *field, const char *delim, int *numfields );
void vsplit( const char *field, const char *delim, char **returnv, int returnl );
char *makefname( char *filename, char *suffix );
int contains( char *string, char *part);

// functions on arrays of numbers
double **histogram( double *data, int size, int bins, int norm);
void isort( int *array, unsigned long len );
void dsort( double *array, unsigned long len );
void isamplewor( int *array, long size, int *subarray, long subsize, long *seed);
void fsamplewor( double *array, long size, double *subarray, long subsize, long *seed);

// other functions
long make_seed();
double gammln( double xx );
double factln( int n );
double bicoln( int n, int k );



/*
 * function: count_lines
 * ---------------------
 * counts number of lines in a file
 */
int count_lines( FILE *fin, char *filename){
	int i;
	int ch, prev = '\n';

	i=0;
	fin = fopen(filename,"rt");
	if( fin == NULL){
		printf("Error in count_lines: I can't open file %s\n", filename);
		exit(0);
	}
	while( (ch = fgetc(fin)) != EOF ){
		if( ch == '\n' ) i++;
		prev = ch;
	}
	fclose(fin);
	if( prev != '\n' ) i++;
	
	return i;
}

/*
 * function: longest_line
 * ----------------------
 * returns longest line in a file (by number of characters)
 */
int longest_line( FILE *fin, char *filename){
	int i, len, oldlen;
	int ch, prev = '\n';

	i=0;
	fin = fopen(filename,"rt");
	if( fin == NULL){
		printf("Error in longest_line: I can't open file %s\n", filename);
		exit(0);
	}
	len=0;
	oldlen=0;
	while( (ch = fgetc(fin)) != EOF ){
		len++;
		if( ch == '\n' ){
			i++;
			if( oldlen < len ) oldlen = len;
			len = 0;
		}
		prev = ch;
	}
	fclose(fin);
	if( prev != '\n' ) i++;
	
	return oldlen+1;
}


/*
 * function: hist (requires alloc.h)
 * ----------------
 * histograms an 1-d input array
 * normalization: 0 = no
 *                1 = sum of bins
 *                2 = integral
 */
double **histogram( double *data, int size, int bins, int norm){
        int i, x;
	double max, min, **hist;
        double normval, dx;

        hist = (double **) alloc2d(sizeof(double), bins, 2);

	/* 
	 * initial conditions, find range of data, count bin widths
	 */
	max=data[0];
	min=data[0];
	for(i=1; i<size; i++){
		if(max < data[i]) max = data[i];
		if(min > data[i]) min = data[i];
	}

	if( norm == 2 )	        // record bin widths if we are normalizing by integral
		dx = (max-min)/bins;
	
	for( i = 0; i < bins ; i++){
		hist[i][1] = 0;
		hist[i][0] = min + (0.5+i)*(max-min)/bins;
	}
	
	/* 
	 * count into bins from data; determine how to normalize from input
	 */
	normval = 1.0;
	
	if( norm == 1 || norm == 2 ){
		normval /=size;                       // normalize to sum of bins
		if( norm == 2)
			normval *= 1.0/dx; // normalize to integral
	}
		
	for( i = 0; i < size; i++){
		x = floor((data[i]-min)*bins/(max-min));
		if( x == bins ) x--;
		hist[x][1]+=1.0*normval;
        }


	return hist;
        
}

/*
 * function: getdir
 * ----------------
 * returns directory of filename
 */
char *getdir( char *filename ){
	int i, j, len;
	char *directory;

	len = strlen( filename );
	directory = (char *) malloc( len*sizeof(char));
	j=0;
	for(i=0; i< len; i++){
		if( filename[i] == '/' ) j=i;
	}
	sprintf(directory,"");
	if( j> 0 ){
		for(i=0; i<= j; i++){
			sprintf(directory,"%s%c", directory, filename[i]);
		}
	}
	
	return directory;
}

/* old version, using strtok
char *getdir( char *filename ){
	int i, j, len;
	char *name, *directory, first;
	
	len = strlen( filename );
	name = (char *) malloc( len*sizeof(char) );
	directory = (char *) malloc( len*sizeof(char) );
	sprintf(name, "%s", filename);
	sprintf(directory, "%s", filename);
	char *field = strtok(name, "/");
	i=0;
	while( field != NULL){
		field = strtok( NULL, "/");
		i++;
	}
	if( filename[0] == '/') sprintf(directory,"/");
	else sprintf(directory,"");
	sprintf(name, "%s", filename);
	field = strtok(name, "/");
	j=0;
	while( j< i-1 && field != NULL){
		sprintf(directory,"%s%s/", directory, field);
		field = strtok( NULL, "/");
		j++;
	}

	char *final = directory;
	return final;
}
*/

/*
 * function: numfields
 * -------------------
 * returns the number of fields in string, delimited by delims
 */
int numfields( char *filename, char *delims ){
	int i, len;
	char *name;
	
	len = strlen( filename );
	name = (char *) malloc( len*sizeof(char) );
	sprintf(name, "%s", filename);
	char *field = strtok(name, delims);
	i=0;
	while( field != NULL){
		field = strtok( NULL, delims);
		i++;
	}

	return i;
}


/*
 * functions: even
 * --------------
 * prints spaces after a given string so that it is of a predefined length, up to a max length
 * if the string is longer than maxlen, then the original string is returned
 *
 * evena evens a string
 * eveni evens an integer
 * evenf evens a float -- ones digit, decimal, and six decimal places means maxlen >= 8
 * evene evens a scientific notation number -- ones digit, decimal, six decimal places, and exponent means maxlen >= 11
 * even evens a string by printing over it (all the others use this)
 *
 * note that a tab is (up to) 8 spaces, if you are using these too
 */
char *evena( char *name, int maxlen ){
	even( name, maxlen);
	return name;
}
			
char *eveni( int val, char *name, int maxlen ){
	name = (char *) malloc( 100*sizeof(char));
	sprintf( name,"%i", val);
	even( name, maxlen);
	return name;
}
			
char *evenf( double val, char *name, int maxlen ){
	name = (char *) malloc( 100*sizeof(char));
	sprintf( name,"%f", val);
	even( name, maxlen);
	return name;
}
			
char *evene( double val, char *name, int maxlen ){
	name = (char *) malloc( 100*sizeof(char));
	sprintf( name,"%e", val);
	even( name, maxlen);
	return name;
}

void even( char *name, int maxlen ){
	int i, slen;
	
	slen = strlen( name );
	
	if( maxlen >= slen ){
		for(i=0; i< maxlen-slen; i++){
			sprintf(name,"%s%s", name, " ");
		}
	}
	else{
		char *new;
		new = (char *) malloc(maxlen*sizeof(char));
		strcpy(new,"");
		for(i=0; i< maxlen; i++)
			sprintf(new,"%s%c", new, name[i]);
		strcpy(name, new);
	}
}

/*
 * function: check_comment
 * -----------------------
 * returns 1 if the provided field begins with "#", 0 otherwise
 */
int check_comment( char *field ){
	if( strlen( field ) > 0 ){
		if( field[0] == '#' ) return 1;
		else return 0;
	}
	else return 0;
}

/*
 * function: get_title
 * -------------------
 * returns the first line of filename in a string
 */
char *get_title( FILE *fin, char *filename ){
	int len;
	char *title;
	int ch, prev = '\n';
	
	len = 0;
	fin = fopen(filename, "rt");
	if( fin == NULL ){
		printf("Error in get_title: I can't open file %s\n", filename);
		exit(0);
	}
	while( (ch = fgetc(fin)) !='\n' ){
		len++;
	}
	fclose(fin);
	len+=2;
	
	title = (char *) malloc( len*sizeof(char));
	
	fin = fopen(filename,"rt");
	fgets( title, len, fin );
	fclose(fin);

	char *final = title;
	return final;
}

/*
 * function: check_ext
 * -------------------
 * checks to see if the string "filename" ends in "ext";
 * returns 1 if it does, 0 if not
 */
int check_ext( char *filename, char *ext ){
	int flen, elen, i, yes;

	flen = strlen( filename );
	elen = strlen( ext );
	if( elen > flen ) return 0;
	else{
		yes = 0;
		for(i=0; i< elen; i++ ){
			if( filename[flen-i-1] == ext[elen-i-1] ) yes++;
		}
		if( yes == elen ) return 1;
		else return 0;
	}
}

/*
 * function: count_nclines
 * -----------------------
 * counts the number of lines that do not contain a comment
 */
int count_nclines( FILE *fin, char *filename){
	int i, newline;
	int ch, prev = '\n';

	i=0;
	fin = fopen(filename,"rt");
	if( fin == NULL){
		printf("Error in count_nclines: I can't open file %s\n", filename);
		exit(0);
	}
	newline = 1;
	while( (ch = fgetc(fin)) != EOF ){
		if( ch == '#' && newline == 1 ) i--;
		if( newline == 1 ) newline = 0;
		if( ch == '\n' ){
			i++;
			newline = 1;
		}
		else newline = 0;
		prev = ch;
	}
	fclose(fin);
	if( prev != '\n' ) i++;
	
	return i;
}
/* old version using fgets
int count_nclines( FILE *fin, char *filename ){
	int i, linel;
	char *line;

	linel = longest_line( fin, filename );
	line = (char *) malloc(linel*sizeof(char));
	
	i=0;
	fin = fopen(filename,"rt");
	if( fin == NULL ){
		printf("Error in count_nclines: I can't open file %s\n", filename);
		exit(0);
	}
	while( fgets( line, linel, fin ) != NULL && !feof(fin) ){
		if( check_comment(line) == 0 ){
			i++;
		}
	}
	fclose(fin);
	free(line);

	return i;
}
*/


/*
 * function: split
 * ---------------
 * splits the field by delim into an array of strings
 */
char **split( const char *field, const char *delim, int *numfields ){
	int i, len, num;
	char *token, *cp, **splits;

	cp = strdup( field );
	len = strlen( field );
	token = strtok( cp, delim );
	i=0;
	while( token != NULL ){
		token = strtok( NULL, delim );
		i++;
	}
	num = i;
	splits = (char **) alloc2d(sizeof(char), num, len);
	strcpy( cp, field);
	token = strtok( cp, delim );
	i=0;
	while( token != NULL ){
		strcpy(splits[i], token);
		token = strtok( NULL, delim );
		i++;
	}
	*numfields = i;
	
	return splits;
}
	
/*
 * function: vsplit
 * ---------------
 * splits the field by delim into an array of strings, up to the length of returnv
 */
void vsplit( const char *field, const char *delim, char **returnv, int returnl ){
	int i, len, num;
	char *token, *cp;

	cp = strdup( field );
	len = strlen( field );
	token = strtok( cp, delim );
	i=0;
	while( token != NULL ){
		token = strtok( NULL, delim );
		i++;
	}
	num = i;
	strcpy( cp, field);
	token = strtok( cp, delim );
	i=0;
	while( token != NULL && i< returnl ){
		strcpy(returnv[i], token);
		token = strtok( NULL, delim );
		i++;
	}
}
	

/*
 * function: isort
 * ---------------
 * sorts an array of integers (using heapsort from numerical methods)
 */
void isort( int *array, unsigned long len ){
	unsigned long i, j, ir, l;
	int rra;
	
	if( len < 2 ) return;
	//l = (len >> 1) +1;
	l = (len/2)+1;
	ir=len;
	for(;;){
		if( l>1 ){
			rra=array[--l-1];
		}
		else{
			rra=array[ir-1];
			array[ir-1]=array[0];
			if( --ir == 1 ){
				array[0] = rra;
				break;
			}
		}
		i=l;
		j=l+l;
		while( j <= ir ){
			if( j< ir && array[j-1] < array[j] ) j++;
			if( rra < array[j-1] ){
				array[i-1]=array[j-1];
				i=j;
				//j <<= 1;
				j*=2;
				
			}
			else j=ir+1;
		}
		array[i-1]=rra;
	}
}
	
/*
 * function: dsort
 * ---------------
 * sorts an array of doubles (using heapsort from numerical methods)
 */
void dsort( double *array, unsigned long len ){
	unsigned long i, j, ir, l;
	double rra;
	
	if( len < 2 ) return;
	//l = (len >> 1) +1;
	l = (len/2)+1;
	ir=len;
	for(;;){
		if( l>1 ){
			rra=array[--l-1];
		}
		else{
			rra=array[ir-1];
			array[ir-1]=array[0];
			if( --ir == 1 ){
				array[0] = rra;
				break;
			}
		}
		i=l;
		j=l+l;
		while( j <= ir ){
			if( j< ir && array[j-1] < array[j] ) j++;
			if( rra < array[j-1] ){
				array[i-1]=array[j-1];
				i=j;
				//j <<= 1;
				j*=2;
				
			}
			else j=ir+1;
		}
		array[i-1]=rra;
	}
}
	

/*
 * function: make_seed
 * -------------------
 * make a random number seed from the date, using time
 * MUST INCLUDE <time.h> in file
 */
long make_seed(){
	long seed;
	int i;
	int wday, yr, mo, day, hr, min, sec;
	time_t now;
	struct tm *ts;
	char buf[80], val[10];

	now = time(NULL);
	ts = localtime(&now);
	strftime(buf, sizeof(buf), "%a %Y %m %d %H %M %S %Z", ts);
	/*
	 * get day of week, write as 1-7 to wday
	 */
	sprintf(val,"%c%c%c", buf[0], buf[1], buf[2]);
	if( strcmp(val, "Mon") == 0 ) wday = 1;
	if( strcmp(val, "Tue") == 0 ) wday = 2;
	if( strcmp(val, "Wed") == 0 ) wday = 3;
	if( strcmp(val, "Thu") == 0 ) wday = 4;
	if( strcmp(val, "Fri") == 0 ) wday = 5;
	if( strcmp(val, "Sat") == 0 ) wday = 6;
	if( strcmp(val, "Sun") == 0 ) wday = 7;
	/*
	 * define other values: year, month, day, hour, minute, second
	 */
	sprintf(val,"%c%c%c%c", buf[4], buf[5], buf[6], buf[7]);
	yr = atoi(val);
	
	sprintf(val,"%c%c", buf[9], buf[10]);
	mo = atoi(val);
	
	sprintf(val,"%c%c", buf[12], buf[13]);
	day = atoi(val);
	
	sprintf(val,"%c%c", buf[15], buf[16]);
	hr = atoi(val);
	
	sprintf(val,"%c%c", buf[18], buf[19]);
	min = atoi(val);
	
	sprintf(val,"%c%c", buf[21], buf[22]);
	sec = atoi(val);
	
	seed = 100000000*sec + 1000000*min + 100000*hr*sec + 10000*day + 100*yr*35*sec + 10*wday + mo;
	if( seed > 0 ) seed *=-1;
	
	return seed;
}

/*
 * function: read_list
 * -------------------
 * reads a list of items from filename (typically other file names) into list
 * adds the string dir to the front (send an empty string if none desired)
 *
 * it checks if the filename contains a space or ends in .xls, if so, I surmise this to be an input from quest,
 * and I separate the files by spaces and add "/" to the front
 */
char **read_list( FILE *fin, char *filename, int *len, char *dir ){
	int i, j, linel, test=0;
	char *val, **list;
	
	for(i=0; i< strlen(filename); i++){
		if( filename[i] == ' ' ) test=1;
	}
	if( check_ext( filename, ".xls" ) == 1 ) test=1;
	if( test!=1 ){
		linel = longest_line(fin, filename)+strlen(dir);
		val = (char *) malloc( linel*sizeof(char));
		*len = count_nclines(fin, filename);
	
		list = (char **) alloc2d( sizeof(char), *len, linel );
		i=0;
		fin = fopen(filename,"rt");
			while( fscanf( fin, "%s", val ) != EOF ){
			sprintf( list[i],"%s%s", dir, val );
			i++;
		}
		fclose(fin);
	}
	else{
		printf("read_lines is inferring your file is from swift, attempting to separate, ");
		*len = 1;
		for(i=0; i< strlen(filename); i++){
			if( filename[i] == ' ' ) (*len)++;
		}
		printf("found %i filenames\n", *len );
		list = (char **) alloc2d( sizeof(char), *len, strlen(filename));
		for(i=0; i< (*len); i++){
			strcpy(list[i],"");
		}
		j=0;
		for(i=0; i< strlen(filename); i++){
			if( filename[i] == ' ' ) j++;
			else sprintf(list[j],"%s%c", list[j], filename[i] );
		}
	}
	
	return list;
}

/*
 * function: makefname
 * -------------------
 * checks if filename has extension suffix, if not, appends suffix
 *
 * also checks if the filename is an unmapped swift file name (contains _concurrent),
 * in which case it does nothing
 */
char *makefname( char *filename, char *suffix ){
	int i, j;
	char *foutname, swift_test[]="_concurrent";
	foutname = (char *) malloc( (strlen(filename)+strlen(suffix)+2)*sizeof(char));
	strcpy( foutname, filename);
	i=0;
	j=0;
	while( i< strlen( filename ) && j < strlen(swift_test) ){
		if( filename[i] == swift_test[j] ){
			j++;
		}
		i++;
	}
	if( check_ext( foutname, suffix ) == 0 && j != strlen(swift_test) ) sprintf(foutname, "%s%s", foutname, suffix);
	return foutname;
}

/*
 * function: contains
 * ------------------
 * checks if string contains part, using strstr
 */
int contains( char *string, char *part){
	if( strstr( string, part ) == NULL ) return 0;
	else return 1;
}

/*
 * function: isamplewor
 * -------------------
 * sampling algorithm w/o replacement for ints
 * can use to shuffle a list by making array and subarray the same size
 */
void isamplewor( int *array, long size, int *subarray, long subsize, long *seed){
	int i, index;
	int tmp;
	long max=size;

	/*
	 * pick a random number between 0 and max,
	 * save that value into subarray,
	 * swap indices of that value and max'th value
	 */
	int j;
	for(i=0; i< subsize; i++){
		index = (int) max*ran1(seed);
		
		tmp = array[index];
		array[index] = array[max-1];
		array[max-1] = tmp;
		subarray[i] = tmp;
		max--;
	}
}

/*
 * function: fsamplewor
 * -------------------
 * sampling algorithm w/o replacement for doubles
 * can use to shuffle a list by making array and subarray the same size
 */
void fsamplewor( double *array, long size, double *subarray, long subsize, long *seed){
	int i, index;
	int tmp;
	long max=size;

	/*
	 * pick a random number between 0 and max,
	 * save that value into subarray,
	 * swap indices of that value and max'th value
	 */
	int j;
	for(i=0; i< subsize; i++){
		index = (int) max*ran1(seed);
		
		tmp = array[index];
		array[index] = array[max-1];
		array[max-1] = tmp;
		subarray[i] = tmp;
		max--;
	}
}

/*
 * function: gammln
 * ----------------
 * log of gamma function (factorial); from numerical recipes
 */
double gammln( double xx ){
	double x, tmp, ser;
	static double cof[6]={76.18009173, -86.50532033, 24.01409822, -1.231739516, 0.120858003e-2, -0.536382e-5};
	int j;

	x = xx-1.0;
	tmp=x+5.5;
	tmp-= (x+0.5)*log(tmp);
	ser=1.0;
	for(j=0; j<= 5; j++){
		x+= 1.0;
		ser+= cof[j]/x;
	}
	return -tmp+log(2.50662827465*ser);
}

/*
 * function: factln
 * ----------------
 * log of the factorial; from numerical recipes
 */
double factln( int n ){
	static double a[101];
	if( n< 0 ){
		printf("Yikes! Don't do a factorial of a negative\n");
		exit(0);
	}
	if( n<= 1) return 0.0;
	if( n<= 100 ) return a[n] ? a[n] : (a[n]=gammln(n+1.0));
	else return gammln(n+1.0);
}

/*
 * function: biocoln
 * -----------------
 * log of binomial coefficient; adapted from numerical recipes
 */
double bicoln( int n, int k ){
	return factln(n)-factln(k)-factln(n-k);
}









